%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%% ============================================================================
%%% Module information

%%% @doc Bridge module of different duplicated code detector algorithms

%%% @author Viktoria Fordos <f-viktoria@elte.hu>

-module(refusr_clone_identifierl).
-vsn("$Rev: 9316 $").
-include("user.hrl").
-include("refusr_clone_identifierl.hrl").

-export([get_clones/0, get_clones/1, default_options/0, error_text/2]).


get_clones()->
    get_clones([]).

get_clones(Options) when is_list(Options) ->
    try
        ExpandedOptions = ?Lib:validate_and_expand_options(?MODULE, Options),
        search(ExpandedOptions)
    catch
        _Cl:E when not is_tuple(E) orelse (tuple_size(E) =/= 3) ->
            ?d({_Cl,E}),
            throw(?LocalErr0r(unexpected_error))
    end.

default_options()->
    [{algorithm, sw_metrics},
     {cache, false}].

error_text(empty_db, _)->
    io_lib:format("Database is empty.~n",[]);
error_text(unexpected_error, _)->
    io_lib:format("Unexpected error occurred.~n",[]).


search(Options)->
    Opt0 = proplists:delete(algorithm, Options),
    OptForAlgs = proplists:delete(cache, Opt0),
    run_selected_alg(proplists:get_value(algorithm, Options),
                     proplists:get_value(cache, Options),
                     OptForAlgs).

run_selected_alg(Algorithm, UseCache, Options)->
    check_common_preconds(),
    run_selected_alg0(Algorithm, select_alg(Algorithm, UseCache),
                      transform_to_subjects(Algorithm, Options)).


run_selected_alg0(matrix, AlgModule, Options)->
    {CandNum, ClonesNum, Clones} =
        lists:foldl(fun(Option, {CandNUmIn, ClonesNumIn, ClonesIn})->
            Result = AlgModule:get_clones(Option),
            ACandNum = proplists:get_value(analysed_candidates_num, Result, 0),
            AClonesNum = proplists:get_value(detected_clones_num, Result, 0),
            AClones = proplists:get_value(detected_clones, Result, []),
            {CandNUmIn + ACandNum, AClonesNum + ClonesNumIn,
             lists:flatten([AClones| ClonesIn])}
        end, {0,0,[]}, Options),
    [{analysed_candidates_num, CandNum},
     {detected_clones_num, ClonesNum},
     {detected_clones, Clones}];
run_selected_alg0(sw_metrics, AlgModule, Options)->
    AlgModule:get_clones(Options).

select_alg(matrix, false)->
    ?TLEClones;
select_alg(matrix, true)->
    ?CachedTLEClones;
select_alg(sw_metrics, _)->
    ?FunClones.


transform_to_subjects(Algorithm, Options)->
    case proplists:get_value(func_list, Options) of
        undefined when Algorithm =:= matrix->
            [Options];
        undefined->
            Options;
        List ->
            Opt0 = proplists:delete(subject, Options),
            Opt = proplists:delete(func_list, Opt0),
            transform_to_subjects0(Algorithm, Opt, List)
    end.

transform_to_subjects0(sw_metrics, Opt0, FunList)->
    [{subject, convert_funs_to_gnodes(func, FunList)} | Opt0];
transform_to_subjects0(matrix, Opt0, FunList)->
    lists:foldl(fun([], OptsIn)->
            OptsIn;
        (Tles, OptsIn)->
            [[{subject, Tles} | Opt0]| OptsIn]
        end,[], convert_funs_to_gnodes(tle, FunList)).


convert_funs_to_gnodes(func, Funs)->
    L = [?Query:exec(?Query:seq(?Mod:find(Mod), ?Fun:find(Fun,Arity)))
            || {Mod, Fun, Arity} <- Funs],
    lists:flatten(L);
convert_funs_to_gnodes(tle, Funs)->
    [?Query:exec(?Query:seq([?Mod:find(Mod), ?Fun:find(Fun,Arity),
                                 ?Fun:definition(), ?Form:clauses(),
                                 ?Clause:body()]))
        || {Mod, Fun, Arity} <- Funs].

check_common_preconds()->
    ?TrueOrThrow(db_not_empty(), ?LocalErr0r(empty_db)).

db_not_empty()->
    length(?Query:exec([file]))>0.

