%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%% ============================================================================
%%% Module information

%%% @doc Library functions of Clone IdentifiErl

%%% @author Viktoria Fordos <f-viktoria@elte.hu>

-module(refusr_clone_identifierl_lib).
-vsn("$Rev: 9316 $").
-include("user.hrl").
-include("refusr_clone_identifierl.hrl").

% interface
-export([format/1, format/2, error_text/2, validate_and_expand_options/2]).
% caching
-export([make_unit_storage/2, free_storage/1, make_match_spec/1]).
% internal
-export([format_worker/2]).

%%% ============================================================================
%%% Result formatting

make_unit_storage(Name, Options)->
    AlphabetOpts = proplists:get_value(alphabet_opts, Options,[]),
    StorageOpts = proplists:get_value(storage_opts, Options,[]),
    AlphaBet = ?Alphabet:convert_graph(AlphabetOpts),
    UnitsWithAlphabet = ?MISC:index_list(AlphaBet),
    Storage = ets:new(Name,
                      [set, {read_concurrency, true}, {keypos, ?unit_key_pos}]
                        ++StorageOpts),
    ets:insert(Storage, UnitsWithAlphabet),
    {Storage, length(UnitsWithAlphabet)}.

free_storage(Storage)->
    ets:delete(Storage).

%% 8> ets:fun2ms(fun({S,N}) when S == 1 orelse S == 2 orelse S == 3-> {N, 2} end).
%% [{{'$1','$2'},
%%   [{'orelse',{'==','$1',1},
%%              {'orelse',{'==','$1',2},{'==','$1',3}}}],
%%   [{{'$2',2}}]}]
%% if record is used
%% ets:fun2ms(fun({#unit{id=Subject},Index}) -> {Subject, Index} end).
%% [{{{unit,'$1','_','_'},'$2'},[],[{{'$1','$2'}}]}]

make_match_spec(Subjects) when length(Subjects) >0 ->
    [{{{unit,'$1','_','_'},'$2'},
      [make_match_spec_cond_rec(Subjects)],
     [{{'$2','$1'}}]}].

make_match_spec_cond_rec([Subject]) ->
    match_spec_eq_cond(Subject);
make_match_spec_cond_rec([Subject1, Subject2]) ->
    {'orelse', match_spec_eq_cond(Subject1), match_spec_eq_cond(Subject2)};
make_match_spec_cond_rec([Subject| Subjects]) ->
    {'orelse', match_spec_eq_cond(Subject), make_match_spec_cond_rec(Subjects)}.

match_spec_eq_cond(Subject) ->
    {'==','$1',{const, Subject}}.
%%% ============================================================================
%%% Result formatting

format(ClonesList)->
    format(ClonesList, [{show_score, false}]).

format(ClonesList, Options)->
    MakeProperArgs = fun(Arg)-> [Arg, Options] end,
    Result = ?MISC:parallelise(ClonesList, ?MODULE, format_worker, MakeProperArgs),
    {_, FormattedClones} = lists:unzip(Result),
    FormattedClones.

format_worker(ClonesList, Options)->
    ScoreFormatter = case proplists:get_value(show_score, Options, false) of
        true ->
            Label = proplists:get_value(score_label, Options, "Similarity: "),
            fun(Score) when is_number(Score)->
                lists:flatten([Label,?MISC:any_to_string(Score),"\n"]);
            (Scores) when is_list(Scores)->
                ScoresStr =
                    string:join(lists:map(fun(S)-> ?MISC:any_to_string(S) end, Scores),", "),
                lists:flatten([Label, ScoresStr,"\n"])
            end;
        _ ->
            fun(_)->
                ""
            end
    end,
    [try
        {L, R, Scores} = unzip_clones(Clone),
        LModName = modulename(hd(L)),
        RModName = modulename(hd(R)),
        delimiter()++
        io_lib:format("Left one (found in ~s):~n",[LModName])++
        format_ent(L)++
        io_lib:format("~nRight one (found in ~s):~n",[RModName])++
        format_ent(R)++io_lib:format("~n",[])++
        ScoreFormatter(Scores)++
        delimiter()++
        io_lib:format("~n~n",[])
    catch
        _:_ -> ""
    end || Clone<-ClonesList, Clone /=[] ].

flat_text(Node = {_,Type, _})
    when Type =:= expr; Type =:= clause; Type =:= form->
    ?Syn:flat_text(Node);
flat_text(Node={_,Type,_}) when Type=:=func ->
    [Form] = ?Query:exec(Node, ?Fun:definition()),
    flat_text(Form).

unzip_clones(#clone_item{fst = LU, snd = RU, score = Sc}) ->
    {[LU], [RU], [Sc]};
unzip_clones(Clone) when is_list(Clone)->
    lists:foldl(
        fun(#clone_item{fst = LU, snd = RU, score = Sc}, {Ls,Rs,Scs}) ->
            {Ls ++ [LU], Rs ++ [RU], Scs ++ [Sc]}
        end, {[],[],[]}, Clone).

modulename(#unit{parent=Clause}) when is_tuple(Clause), element(2,Clause) == clause->
    [Node] = ?Query:exec(Clause, ?Query:seq(?Clause:form(),?Form:module())),
    io_lib:format("~p", [?Mod:name(Node)]);
modulename(#unit{parent=Form}) when is_tuple(Form), element(2,Form) == form->
    [Node] = ?Query:exec(Form, ?Form:module()),
    io_lib:format("~p", [?Mod:name(Node)]);
modulename(#unit{parent=Module}) when is_tuple(Module), element(2,Module) == module->
    io_lib:format("~p", [?Mod:name(Module)]).

format_ent(List)->
    S=string:join([flat_text(Node) || #unit{id = Node} <- List], ",\n"),
    string:sub_string(S,string:span(S, " \t\n")+1).

delimiter()->
    string:concat(string:copies("-",80),"\n").

%%% ============================================================================
%%% Error handling

error_text(badarg, [Key, Val])->
    io_lib:format("Bad value (~p) was given for an argument (~p)~n.",[Val, Key]);
error_text(badarg, [Key])->
    io_lib:format("Bad value was given for the following argument: ~p~n.",[Key]);
error_text(badarg, [])->
    io_lib:format("Bad argument was given. ~n.",[]).


%%% ============================================================================
%%% Input validation
validate_and_expand_options(Module, Options)->
    Defaults = Module:default_options(),
    Expanded = expand_options(Options, Defaults),
    validate_options(Module, Expanded),
    Expanded.

expand_options(Options, Defaults)->
    lists:foldl(fun({Key, Val}, OptsIn) ->
        case proplists:get_value(Key, OptsIn) of
            undefined -> [{Key, Val} | OptsIn];
            _ -> OptsIn
        end
    end, Options, Defaults).

badarg(Key)->
    ?LocalError(badarg, [Key]).
%badarg(Key, Val)->
%    ?LocalError(badarg, [Key, Val]).

validate_options(?CloneIdentifiErl, Options) when is_list(Options)->
    MandatoryOptKeys = [algorithm, cache],
    OptionalOptKeys = [func_list],
    lists:all(fun(Key)->
        ?TrueOrThrow(validate(Key, proplists:get_value(Key, Options, undef)),
            badarg(Key))
              end, MandatoryOptKeys)  andalso
        lists:all(fun(Key)->
            ?TrueOrThrow(((not proplists:is_defined(Key, Options)) orelse
                validate(Key, proplists:get_value(Key, Options, undef))),
                badarg(Key))
                  end, OptionalOptKeys) orelse throw(?LocalErr0r(badarg));

validate_options(?TLEClones, Options) when is_list(Options)->
    MandatoryOptKeys = [subject, metric, method, diff_limit, max_invalid_seq_length],
    OptionalOptKeys = [alphabet_opts],
    lists:all(fun(Key)->
        ?TrueOrThrow(validate(Key, proplists:get_value(Key, Options, undef)),
            badarg(Key))
              end, MandatoryOptKeys) andalso
        lists:all(fun(Key)->
            ?TrueOrThrow(((not proplists:is_defined(Key, Options)) orelse
                             validate(Key, proplists:get_value(Key, Options, undef))),
                badarg(Key))
                  end, OptionalOptKeys) orelse
        throw(?LocalErr0r(badarg));

validate_options(?FunClones, Options) when is_list(Options)->
    OptionalOptKeys = [subject, unit, max_rank],
        lists:all(fun(Key)->
            ?TrueOrThrow(((not proplists:is_defined(Key, Options)) orelse
                             validate(Key, proplists:get_value(Key, Options, undef))),
                badarg(Key))
                  end, OptionalOptKeys) orelse
        throw(?LocalErr0r(badarg));

validate_options(?Alphabet, Opts) when is_list(Opts)->
    NullArityFuns = [file_filter, form_filter, clause_filter,
                     expr_filter, token_filter, alphabet],
    OneArityFuns = [tokenizer],
    ValidUnits = [file, form, clause, expr, token],
    ValidTypes = [to_alphabet, to_unit, to_mod_and_fun_preserving_alphabet,
                  to_fun_preserving_alphabet],
    (?TrueOrThrow(lists:member(proplists:get_value(unit, Opts, undef), ValidUnits),
        badarg(unit)) andalso
         ?TrueOrThrow(lists:member(proplists:get_value(type, Opts, undef), ValidTypes),
            badarg(type)) andalso
         lists:all(fun(Key)->
            ?TrueOrThrow(is_function(proplists:get_value(Key, Opts, fun(_)->ok end), 1),
                badarg(Key))
                   end, OneArityFuns) andalso
         lists:all(fun(Key)->
            ?TrueOrThrow(is_function(proplists:get_value(Key, Opts, fun()->ok end),0),
                badarg(Key))
                   end, NullArityFuns))  orelse
        throw(?LocalErr0r(badarg));

validate_options(_,_)->
    throw(?LocalErr0r(badarg)).

validate(func_list, [])->
  false;
validate(func_list, Funs) when is_list(Funs)->
  lists:all(fun(E) when is_tuple(E)->
                    Items = tuple_to_list(E),
                    [Mod, Fun, Arity |_] = Items,
                    (length(Items) == 3) andalso
                        lists:all(fun erlang:is_atom/1, [Mod, Fun]) andalso
                        is_integer(Arity) andalso (Arity >= 0);
                (_)->
                     false
        end, Funs);
validate(func_list, _)->
  false;
validate(cache, Val) ->
  lists:member(Val, [false, true]);
validate(algorithm, Algorithm)->
   lists:member(Algorithm, [sw_metrics, matrix]);
validate(subject, [{'$gn', _ , _ }|_])->
    true;
validate(subject, Subject) ->
    Subject =:= all;
validate(metric, Metric) ->
    (Metric =:= leveinstein) orelse (Metric =:= dice_sorensen)
        orelse is_function(Metric, 2);
validate(method, Method) ->
    (Method =:= quick) orelse (Method =:= standard) orelse (Method =:= precalc);
validate(diff_limit, Limit) ->
    (Limit >= 0) andalso (Limit =< 1);
validate(max_invalid_seq_length, Length) ->
    is_integer(Length) andalso (Length >= 0);
validate(alphabet_opts, AOpts) ->
    is_list(AOpts);
validate(unit, Unit) ->
    lists:member(Unit, [form, clause]);
validate(max_rank, MaxRank) when is_integer(MaxRank)->
    (MaxRank >= 1) andalso (MaxRank =< 5);
validate(_,_)->
    false.
