%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Lilla Hajs <lya@elte.hu>

-module(refusr_sq).
-vsn("$Rev: 10849 $ ").

-export([run/3]).
-export([format_nodes/2]).
-export([prepare/1, error_text/2]).
-export([closure_worker/4, chains_worker/5, pmap_helper/2]).

-include("user.hrl").
-include("sq_lib.hrl").

-define(Lib, refusr_sq_lib).
-define(Format, refusr_sq_format).


%%% ============================================================================
%%% Errors

error_text(lexical_error, Error) ->
    refusr_sq_lexer:format_error(Error);
error_text(syntax_error, Error) ->
    refusr_ac_parser:format_error(Error);
error_text(illegal_property, Params) ->
    io_lib:format("illegal ~p property: ~p", Params);
error_text(illegal_selector, Params) ->
    io_lib:format("illegal ~p selector: ~p  ", Params);
error_text(type_mismatch, [At, Of, ExpType]) ->
    [ case At of
          {Pl,De} -> io_lib:format("type mismatch at ~p ~p :\n    ", [Pl,De]);
          Pl -> io_lib:format("type mismatch at ~p:\n    ", [Pl])
      end,
      case Of of
          {Ent, Type} -> io_lib:format("the type of ~p is ~s, not ~s",
                                       [Ent, Type, ExpType]);
          Ent -> io_lib:format("the type of ~p is not ~s", [Ent, ExpType])
      end ];
error_text(no_property_in_comparison, []) ->
    "no property in comparison";
error_text(statistics_error, []) ->
    "statistics are only available for properties with numeric values";
error_text(bad_regexp, [])->
    "illegal regexp".
%%% ============================================================================
%%% Callbacks

%% @spec run(DisplayOpt::proplist(), Params::proplist(), Query::string()) ->
%%           QueryResult::term()
%% @doc Returns the result of `Query' starting from the initial state given by
%%      `Params'. The format of the result is determined by `DisplayOpt'.
%%
%%      `Params' contains either the
%%          - (optional) `file' and `position' keys or
%%          - the `node_list' and (optional) `node_type' key.
%%       The possible values for
%%          - `node_type': file|function|record|field|macro|variable|expression
%%
%%      `DisplayOpt' contains the keys `positions' and `output'.
%%      The possible values for
%%          - `positions': none|scalar|linecol|both
%%          - `output': stdio|{iodev,io_device()}|msg|other|nodes
%%
%%      The `QueryResult' depends on the `output' key in `DisplayOpt'.
%%          - stdio: a formatted text written to stdio
%%          - {iodev,Dev::io_device()}: a formatted text written to Dev
%%          - msg: a message containing a list with the following types of
%%                 elements: {eq, Name, Value} |
%%                           {list, [{Position, Text}]} |
%%                           {chain, [{Position, Text}], PostWS} |
%%                           {group_by, {Position, Text}, eq, Name, Value} |
%%                           {group_by, {Position, Text},
%%                            list, [{Position, Text}]}
%%          - other: the same list that is otherwise sent by a message
%%          - nodes: a proplist with the keys: `nodes' for a list of nodes and
%%                                             `text' for a formatted text
%%      The format of positions depends on the `positions' key in `DisplayOpt'.
%%          - none: nopos
%%          - scalar: {File::string(), PosFrom::integer(), PosTo::integer()}
%%          - linecol: {File::string(), PosFrom::{integer(), integer()},
%%                                      PosTo::{integer(), integer()}}
run(DisplayOpt, Params, Query) when is_list(Query) ->
    Tokens   = tokenize(Query),
    SynTree  = parse(Tokens),
    SynTree2 = make_internal_representation(SynTree),
    Result   = process_semantic_query(Params, SynTree2),
    ?Format:result(Result, DisplayOpt).

make_internal_representation(SynTree) ->
    case refusr_ac:validate_query(SynTree) of
        {ok, Res} ->
            Res;
        {error, E} ->
            throw(?LocalError(syntax_error, E))
    end.

parse(Tokens) ->
    case refusr_ac_parser:parse(Tokens) of
        {ok, SynTree} ->
            SynTree;
        {error, {_, _, Err}} ->
            case lists:flatten(Err) of
                "{quoted," ++ A ->
                    A1 = lists:reverse(tl(lists:reverse(A))),
                    throw(?LocalError(syntax_error, A1));
                _ ->
                    throw(?LocalError(syntax_error, Err))
            end
    end.

tokenize(Query) ->
    case refusr_sq_lexer:string(Query) of
        {ok, Tokens, _} ->
            Tokens;
        {error, {_, _, Error}, _} ->
            throw(?LocalError(lexical_error, Error))
    end.

%% @private
%% @todo default? think through what's needed + not sure if eclipse needs this
prepare(Args) ->
    fun () ->
        [DisplayOpt,StartOpt,QueryStr] =
            ?MISC:pgetu([display_opt,start_opt,querystr],Args),
        run(DisplayOpt,
            StartOpt,
            QueryStr)
    end.

%%% ============================================================================
%%% Implementation

semantic_query_last_type(SemanticQuery) ->
    [{initial_selector, InitialSelector}| Seq] = SemanticQuery,
    Istate = fun({quoted, InitSel}) ->
                 ?Lib:init_sel_type(InitSel);
                (InitSel) ->
                 ?Lib:init_sel_type(InitSel)
             end,
    InitType = Istate(InitialSelector),

    LastState = lists:foldl(fun check/2, #state{type=InitType}, Seq),

    LastState#state.type.

process_semantic_query(Params, SemanticQuery) ->
    {InitialSelector, _, Seq, Last} =
    lists:foldr(
      fun(Filter = {filter, _F}, {undef, undef, [], Last}) ->
          {undef, undef, [], [Filter|Last]};
         ({initial_selector, InitialSelector}, {undef, Type, Mid, Last}) ->
          {InitialSelector, Type, Mid, Last};
         (Elem, {undef, undef, [], Last}) ->
          {undef, element(1,Elem), [], [Elem|Last]};
         (Filter = {filter, _F}, {undef, Type, Mid, Last}) ->
          {undef, Type, [Filter|Mid], Last};
         (Closure = {closure, _S, _M}, {undef, closure, Mid, Last}) ->
          {undef, closure, [], [Closure|Mid++Last]};
         (Iteration = {iteration, _S, _M}, {undef, iteration, Mid, Last}) ->
          {undef, iteration, [], [Iteration|Mid++Last]};
         (Elem, {undef, _Type, Mid, Last}) ->
          {undef, mid, [Elem|Mid], Last}
      end,
      {undef, undef, [], []},
      SemanticQuery),

    InitState = initial_state(#state{params=Params}, InitialSelector),
    SeqState = lists:foldl(fun check/2, InitState, Seq),
    LastState = lists:foldl(fun check/2,
                            SeqState#state{checked_query=[]}, Last),
    case {LastState#state.type, LastState#state.checked_query} of

        {help, {EntityType, HelpType}} ->
            {help, {EntityType, HelpType}};
        _ ->
            State =
            process_query_seq(InitState, SeqState#state.checked_query),
            process_last_query(State, LastState#state.checked_query)
    end.

process_query_seq(State,QuerySeq) ->
    lists:foldr(fun process/2, State, QuerySeq).

process_last_query(State, LastQuery) ->
    lists:foldr(fun process_last/2, State, LastQuery).

%%% ============================================================================
%%% Preprocessing of queries

%todo:  node list <-> query format check
initial_state(State, {quoted, InitialSelector}) ->
    initial_state(State, InitialSelector);
initial_state(#state{params=Params}=State, InitialSelector) ->
    {NodeType, NodeList} =
        case proplists:get_value(node_list, Params) of
            undefined ->
                case InitialSelector of
                    undef -> {none, []};
                    _ ->
                        ?Lib:init_sel([{ask_missing,false} |Params],
                                      InitialSelector)
                end;
            [] ->
                {none, []};
            [HNode|TNodes] = Nodes ->
                Type = proplists:get_value(node_type, Params,
                                           ?Lib:node_type(HNode)),

                NT = [{Node, ?Lib:node_type(Node)}|| Node <- TNodes],
                Diffs = lists:filter(fun({_N, NType}) -> NType/=Type end, NT),
                case Diffs of
                    [] ->
                        {Type, Nodes};
                    _ ->
                        Params = ["initial state", hd(Diffs), Type],
                        throw(?LocalError(type_mismatch, Params))
                end
        end,
    State#state{type=NodeType, res=ordsets:from_list(NodeList)}.


check({selector, Sel}, #state{type=Type, checked_query=Lst}=State) ->
    Selector =
        case Sel of
            {quoted, Atom} -> Atom;
            Atom -> Atom
        end,
    case ?Lib:sel_type(Type, Selector) of
        [SelType] ->
            State#state{type=SelType,
                        checked_query=[{selector, Selector, SelType}| Lst]};
        [] ->
            case ?Lib:prop_type(Type, Selector) of
                [PropType] ->
                    State#state{type = PropType,
                                checked_query =
                                    [{property, Selector, PropType}| Lst]};
                [] -> throw(?LocalError(illegal_selector, [Type, Selector]))
            end
    end;

check({Action, {seq, Seq}, {mult, Mult}},
        #state{type=Type, checked_query=Lst}=State) ->
    case lists:foldl(fun check/2, State#state{checked_query = []}, Seq) of
        #state{type=Type, checked_query=QSLst} ->
            case Mult of
                0 -> State#state{checked_query = Lst};
                _ -> State#state{checked_query = [{Action, QSLst, Mult}| Lst]}
            end;
        #state{type=BadType} ->
            throw(?LocalError(type_mismatch, [Action, {Seq, BadType}, Type]))
    end;

check({filter, Filter}, #state{checked_query=Lst}=State) ->
    State#state{checked_query = [{filter, check_filter(State, Filter)}|Lst]};

check({statistics, Stat}, #state{type=PropType}=State) ->
    Statistics =
        case Stat of
            {quoted, S} -> S;
            S           -> S
        end,
    case PropType of
        any ->
            State#state{type=int, checked_query = [{statistics, Statistics}]};
        int ->
            State#state{type=int, checked_query = [{statistics, Statistics}]};
        _ ->
        throw(?LocalError(statistics_error, []))
    end;

%todo: filterekben help?
check({help, HelpType}, #state{type=Type}=State) ->
    State#state{type=help, checked_query=
        case HelpType of
            initial_selectors -> {initial_selectors, []};
            queries           -> {selectors, Type};
            statistics        -> {statistics, []};
            filters           -> {properties, Type}
        end
    }.


%% @private
%% @spec check_filter(Type::atom(), Filter::atom()|tuple()) -> atom()|tuple()
check_filter(_State, 'true') ->
    'true';

check_filter(_State, 'false') ->
    'false';

check_filter(State, {'or', Filter1, Filter2}) ->
    {'or', check_filter(State, Filter1), check_filter(State, Filter2)};

check_filter(State, {'and', Filter1, Filter2}) ->
    {'and', check_filter(State, Filter1), check_filter(State, Filter2)};

check_filter(State, {'not', Filter}) ->
    {'not', check_filter(State, Filter)};

check_filter(State,{'in', {seq,SelList}, List})->
    Query = lists:foldl(fun check/2, State#state{checked_query=[]}, SelList),
    case List of
        {ok,SemQuery} ->
            QType = Query#state.type,
            SQType = semantic_query_last_type(SemQuery),
            ?Check(SQType == QType orelse
                   ((QType==atom orelse SQType==atom) andalso
                    (QType==string orelse SQType==string)),
                ?LocalError(type_mismatch,
                            [{filter, {'in', SelList, List}}, {List}, QType]));
        _ -> ok
    end,
    {'in', {seq, Query#state.checked_query}, List };

check_filter(#state{type=Type} = State,{'in', Property, List})->
    case List of
        {cons,L} ->
            lists:map(fun (E)->check_filter(State,{'==',Property,E}) end,L),
            {'in', Property, {cons,L}};
        []->
            {'in', Property, {cons,[]}};
        {ok, Query} ->
            PType = hd(?Lib:prop_type(Type, Property)),
            SQType = semantic_query_last_type(Query),
            ?Check(SQType == PType orelse
                   ((PType==atom orelse SQType==atom) andalso
                    (PType==string orelse SQType==string)),
                ?LocalError(type_mismatch,
                            [{filter, {'in', Property, List}},
                                {List}, PType])),
            {'in', Property, {'query',Query }}
    end;

check_filter('in',{Type,Filter1,{_,_,_}=Filter2}) ->
    ?Check( (Filter1 == ?Lib:node_type(Filter2)) orelse
            (Filter1 == module andalso (?Lib:node_type(Filter2)==file)),
               ?LocalError(type_mismatch,
                           [{filter, {'in', Type, Filter2}},
                                Filter2, Filter1]));

check_filter('in',{Type,Filter1,Filter2})->
    ?Check(Filter1 == any orelse
            (Filter1 == int andalso is_integer(Filter2)) orelse
            (Filter1 == atom andalso is_atom(Filter2)) orelse
            (Filter1 == string andalso is_list(Filter2)) orelse
            (Filter1 == atom andalso is_list(Filter2)) orelse
            (Filter1 == string andalso is_atom(Filter2)) orelse
            (Filter1 == bool andalso is_boolean(Filter2)),
            ?LocalError(type_mismatch,
                       [{filter, {'in', Type, Filter2}},
                        Filter2, Filter1]));

check_filter(State, {seq, Seq}) ->
    St = lists:foldl(fun check/2, State#state{checked_query=[]}, Seq),
    {seq, St#state.checked_query};

check_filter(_State, {_CompOp, {quoted, _}, {quoted, _}}) ->
    throw(?LocalError(no_property_in_comparison, []));

check_filter(State, {CompOp, {quoted, _} = F1, F2}) ->
    check_filter(State, {CompOp, F2, F1});

check_filter(#state{type=Type}=State, {CompOp, F1, {quoted, _} = F2}) ->
    if
        is_tuple(F1) -> {CompOp, check_filter(State, F1), F2};
        is_atom(F1)  ->
            PropType1 = ?Lib:prop_type(Type, F1),
            ?Check(PropType1 =:= [atom],
                   ?LocalError(no_property_in_comparison, [])),
            {CompOp, F1, F2};
        true ->
            throw(?LocalError(no_property_in_comparison, []))
    end;

check_filter(State, {CompOp, F1, F2}) when is_tuple(F1) andalso is_tuple(F2) ->
    {CompOp, check_filter(State, F1), check_filter(State, F1)};

check_filter(#state{type=Type}=State, {CompOp, F1, F2}) when is_tuple(F1) ->
    ?Check(?Lib:prop_type(Type, F2) == [bool] orelse is_boolean(F2),
           ?LocalError(type_mismatch, [{filter,{CompOp, F1, F2}}, F2, bool])),
    {CompOp, check_filter(State, F1), F2};

check_filter(State, {CompOp, F1, F2}) when is_tuple(F2) ->
    {CompOp, CheckedF2, F1} = check_filter(State, {CompOp, F2, F1}),
    {CompOp, F1, CheckedF2};

check_filter(#state{type=Type}, {CompOp, name, F2}) ->
    TypeOfName = ?Lib:prop_type(Type, name),
    %`Type' doesn't have a `name' property
    ?Check(TypeOfName /= [], ?LocalError(illegal_property, [Type, name])),

    PropTypeOfF2 = ?Lib:prop_type(Type, F2),

    case PropTypeOfF2 of
        %%`F2' is not a property, but not string neither atom
        [] when not (is_list(F2) orelse is_atom(F2)) ->
            throw(?LocalError(type_mismatch, [{filter, {CompOp, name, F2}},
                                              F2, hd(TypeOfName)]));

        %%`F2' is a property, but it differs in type
        [_] when TypeOfName /= PropTypeOfF2 ->
            throw(?LocalError(type_mismatch, [{filter, {CompOp, name, F2}},
                                              {F2, hd(PropTypeOfF2)},
                                              hd(TypeOfName)]));

        %%`F2' is a property and the two types are matching, or is not
        %%a property, but is a type of atom or string
        _ -> ok

    end,

    %% %`F2' is a property, but not of a matching type
    ?Check(PropTypeOfF2 == [] orelse TypeOfName == PropTypeOfF2,
           ?LocalError(type_mismatch,
                        [{filter, {CompOp, name, F2}},
                            {F2, hd(PropTypeOfF2)},hd(TypeOfName)])),
    %%`F2' is not a property, but not string or atom either
     ?Check(PropTypeOfF2 == [] andalso (is_list(F2) orelse is_atom(F2)),
            ?LocalError(type_mismatch, [{filter, {CompOp, name, F2}},
                                        F2, hd(TypeOfName)])),
    TypeOfF2 = if
                   PropTypeOfF2 /= [] ->PropTypeOfF2;
                   is_list(F2)        -> [string];
                   is_atom(F2)        -> [atom]
               end,

    if
        TypeOfName == TypeOfF2 -> {CompOp, name, F2};
        TypeOfF2 == [atom]     -> {CompOp, name, atom_to_list(F2)};
        TypeOfF2 == [string]   -> {CompOp, name, list_to_atom(F2)}
    end;

check_filter(State, {CompOp, F1, name}) ->
    {CompOp, name, Comp2} = check_filter(State, {CompOp, name, F1}),
    {CompOp, Comp2, name};

check_filter(#state{type=Type}, {CompOp, F1, F2})
                when is_atom(F1) andalso is_atom(F2)->
    PropType1 = ?Lib:prop_type(Type, F1),
    PropType2 = ?Lib:prop_type(Type, F2),
    ?Check(PropType1 /= [] orelse PropType2 /= [],
           ?LocalError(no_property_in_comparison, [])),

    Type1 = case PropType1 of [] -> [atom]; _ -> PropType1 end,
    Type2 = case PropType2 of [] -> [atom]; _ -> PropType2 end,

    if
        Type1 == Type2 -> {CompOp, F1, F2};
        Type1 == [any] orelse Type2 == [any] -> {CompOp, F1, F2};
        Type1 == [bool] andalso is_boolean(F2) -> {CompOp, F1, F2};
        Type2 == [bool] andalso is_boolean(F1) -> {CompOp, F1, F2};
        true -> throw(?LocalError(type_mismatch, [{filter, {CompOp, F1, F2}},
                                                  {F1, hd(Type1)}, hd(Type2)]))
    end;

check_filter(#state{type=Type}, {CompOp, F1, F2}) when is_atom(F1) ->
    PropType = ?Lib:prop_type(Type, F1),
    ?Check(PropType /= [],
           ?LocalError(no_property_in_comparison, [])),

    ?Check(PropType == [any] orelse
              (PropType == [int] andalso is_integer(F2)) orelse
              (PropType == [atom] andalso is_atom(F2)) orelse
              (PropType == [string] andalso is_list(F2)) orelse
              (PropType == [bool] andalso is_boolean(F2)),
           ?LocalError(type_mismatch,
                       [{filter, {CompOp, F1, F2}}, F2, hd(PropType)])),

    {CompOp, F1, F2};

check_filter(#state{type=Type}, {CompOp, F1, F2}) when is_atom(F2) ->
    check_filter(Type, {CompOp, F2, F1}),
    check_filter(Type, {CompOp, F2, F1}),
    {CompOp, F1, F2};

check_filter(_State, {_CompOp, _F1, _F2}) ->
    throw(?LocalError(no_property_in_comparison, []));

check_filter(#state{type=Type}, Filter) ->
    PropType = ?Lib:prop_type(Type, Filter),
    case PropType of
        [bool] ->
            Filter;
        [] ->
            throw(?LocalError(illegal_property, [Type, Filter]));
        _ ->
            throw(?LocalError(type_mismatch,
                              [{filter,Filter}, {Filter, hd(PropType)}, bool]))
    end.
%%% ============================================================================
%%% Processing of queries

process({selector, Selector, SelType},
        #state{type= Type, res= Entities, params=Params}) ->
    [Fun] = ?Lib:sel_fun(Type, Selector),
    #state{action = selection,
           type   = SelType,
           res    = ordsets:fold(
                       fun(Entity, Acc) ->
                           ordsets:union(ordsets:from_list(Fun(Entity)), Acc)
                       end,
                       ordsets:new(),
                       Entities),
           params    = Params,
           prev_type = Type,
           prev_res  = Entities,
           groupby_type = Type,
           groupby_res = Entities};

process({_Action, _QSLst, 0}, State) -> State;

process({iteration, QSLst, Mult}, State) ->
    process({iteration, QSLst, Mult-1}, process_query_seq(State, QSLst));

process({closure, QSLst, Mult}, #state{res=Entities}=St) ->
    Tab  = ets:new(store, [public, ordered_set, {write_concurrency, true}]),
    EntityLists = group_entities(Entities),
    [ets:insert(Tab, {E}) || E <- Entities],
    Keys = [ rpc:async_call(node(), ?MODULE, closure_worker,
                            [St#state{res=[E]},QSLst, Mult, Tab])
             || E <- EntityLists],
    [ ok = rpc:yield(K) || K <- Keys],
    Result =
        ets:select(Tab, [{{'$1'},[], ['$1']}]),
    ets:delete(Tab),
    St#state{action=closure,
          res=Result};


process({filter, Filter}, #state{action=selection, type=Type, res=Res}=St) ->
    St#state{
          res=filter(Filter, St),
           prev_type = Type,
           prev_res = Res,
           groupby_type = Type,
           groupby_res = Res};



process({property, Prop, PropType},
        #state{type=Type, res=Entities,params=Params}) ->

    #state{action    = property_query,
           type = PropType,
           res       = pmap({?MODULE, pmap_helper},
                                ?Lib:prop_fun(Type, Prop), Entities),
           params = Params,
           prev_type = Type,
           prev_res = Entities,
           groupby_type = Type,
           groupby_res = Entities}.

group_entities(Entities) ->
    NumberOfData = length(Entities),
    Divide =
        case erlang:system_info(logical_processors_available) of
            unknown ->
                (NumberOfData div 10) + 1;
            N ->
                (NumberOfData div (N*10)) + 1
        end,
    ?MISC:slice_list(Divide, Entities).

%% @private
closure_worker(_State,_QSLst, 0, _Tab) ->
    ok;
closure_worker(#state{res=[]},_QSLst, _Mult, _Tab) ->
    ok;
closure_worker(#state{type=Type}=State,QSLst, Mult, Tab) ->
    #state{type=Type, res=NewEntities} = process_query_seq(State, QSLst),
    ReallyNewEntities =
        [begin
             ets:insert(Tab, {NEntity}),
             NEntity
         end || NEntity <- NewEntities, not  ets:member(Tab, NEntity)],
    NewMult = case Mult of infinite = I -> I; M -> M-1 end,
    closure_worker(State#state{res=ReallyNewEntities},QSLst, NewMult, Tab).


process_last(_Query, #state{res=[]}=State) -> State#state{prev_res=[]};

process_last({selector, Selector, SelType},
                #state{type=Type, res=Entities,params=Params}) ->
    #state{action    = selection, %TODO: set instead of list?
           type      = SelType,
           res       = pmap({?MODULE, pmap_helper},
                                ?Lib:sel_fun(Type, Selector), Entities),
           params    = Params,
           prev_type = Type,
           prev_res = Entities};

process_last({filter, Filt},#state{action=selection, prev_res=[]}=St) ->
    St#state{res=filter(Filt, St)};

process_last({filter, Filt}, #state{action=selection, res=Res}=St) ->
    St#state{res=[filter(Filt, St#state{res= Entities})|| Entities <- Res]};

process_last({property, Prop, _PropType},
                #state{type=Type, res=Entities,params=Params}) ->
    #state{action    = property_query,
           type      = Prop,
           res       = pmap({?MODULE, pmap_helper},
                                ?Lib:prop_fun(Type, Prop), Entities),
           params = Params,
           prev_type = Type,
           prev_res = Entities};

process_last({Action, QSLst, Mult},
                #state{type=Type, res=Res,params=Params}=State) ->
    InitialChains =
        case State#state.action of
            selection -> #chains{incomplete=[[Entity]|| Entity<-Res]};
            Action -> Res
        end,
    #state{action = Action,
            type = Type,
            res = chains(State#state{action=Action},
                            InitialChains, QSLst, Mult),
            params = Params};

process_last({filter, Filt}, #state{action=iteration, res=Res}=St) ->
    NewIncomplete = lists:filter(
                      fun(Chain) ->
                        filter(Filt, St#state{res= [hd(Chain)]}) /= [] end,
                      Res#chains.incomplete),

    St#state{res = Res#chains{incomplete = NewIncomplete}};

process_last({filter, Filt}, #state{action=closure, res=Res}=St) ->
    NewComplete = filter_chain(St, Filt, Res#chains.complete),
    NewIncomplete = filter_chain(St, Filt, Res#chains.incomplete),
    NewRecursive = filter_chain(St, Filt, Res#chains.recursive),

    NewChains = Res#chains{complete   = NewComplete,
                           incomplete = NewIncomplete,
                           recursive  = NewRecursive},

    St#state{res = NewChains};

% todo: preproc
process_last({statistics, Stat}, #state{res=PropValues}) ->
    NonNum = lists:filter(fun(Val) -> not is_number(Val) end, PropValues),

    ?Check(NonNum == [],
            ?LocalError(type_mismatch,
                        [{statistics,Stat}, hd(NonNum), number])),

    Fun = ?Lib:stat_fun(Stat),
    #state{action = statistics, type = Stat, res = Fun(PropValues)}.

%%% ============================================================================
%%% Helper functions
pmap({Module, Func}, ExtraArgs, List) ->
    % rpc:pmap({Module, Func}, ExtraArgs, List). %slower than ?MISC:parallelise
     ?MISC:parallelise(List, Module, Func,
                       fun(Entities)-> [Entities|ExtraArgs] end, false).

pmap_helper(Entities, Fun) when is_list(Entities) ->
     [Fun(Entity) || Entity <- Entities];
pmap_helper(Entity, Fun)->
     pmap_helper([Entity], Fun).

chains_worker(_State, [] = _Chains, _QSLst, _Mult, _Tab) ->
    ok;
chains_worker(_State, Chains, _QSLst, 0, Tab) ->
    ets:insert(Tab, [{incomplete, Chain} || Chain <- Chains]),
    ok;
chains_worker(#state{type=Type}=State,Chains, QSLst, Mult, Tab) ->
    NewChains =
        lists:flatmap(
          fun([H|_] = Chain) ->
                #state{type=Type, res=Res} =
                    process_query_seq(State#state{res = [H]}, QSLst),
                next_chain(State#state{res=Res}, Chain)
          end,
          Chains),
    SplitFun =
        fun({complete, _} = C, {Incomp, Comp, Rec}) ->
                {Incomp, [C | Comp], Rec};
           ({recursive, _} = R, {Incomp, Comp, Rec}) ->
                {Incomp, Comp, [R | Rec]};
           (List, {Incomp, Comp, Rec}) ->
                {[List | Incomp], Comp, Rec}
        end,

    {Incomplete, Completed, Recursive} =
        lists:foldl(SplitFun, {[], [], []}, NewChains),

    ets:insert(Tab, Completed ++ Recursive),

    NewMult = case Mult of infinite -> infinite; Mult -> Mult-1 end,
    chains_worker(State,Incomplete, QSLst, NewMult, Tab).

chains(_State,#chains{incomplete = []} = Chains, _QSLst, _Mult) ->
    Chains;
chains(State,#chains{incomplete = Chains} = ChainsRec, QSLst, Mult) ->
    Tab  = ets:new(store, [public, bag, {write_concurrency, true}]),
    ChainGroups = group_entities(Chains),
    Keys = [rpc:async_call(node(), ?MODULE, chains_worker,
                           [State, Ch, QSLst, Mult, Tab])
            || Ch <- ChainGroups],
    [ ok = rpc:yield(K) || K <- Keys],

    Recursive =
        ets:select(Tab, [{{'$1', '$2'},[{'==', '$1', recursive}], ['$2']}]),
    Completed =
        ets:select(Tab, [{{'$1', '$2'},[{'==', '$1', complete}], ['$2']}]),
    Incomplete =
        ets:select(Tab, [{{'$1', '$2'},[{'==', '$1', incomplete}], ['$2']}]),
    ets:delete(Tab),
    ChainsRec#chains{complete   = Completed ++ ChainsRec#chains.complete,
                     incomplete = Incomplete,
                     recursive  =  Recursive ++ ChainsRec#chains.recursive}.

next_chain(#state{action=iteration, res = Result}, Chain) ->
    lists:foldl(fun(Entity, Acc) -> [[Entity| Chain]|Acc] end, [], Result);
next_chain(#state{action=closure, res=[]}, Chain) ->
    [{complete, Chain}];
next_chain(#state{action=closure, res=Result}, Chain) ->
    lists:foldl(
      fun(Entity, Acc) ->
              case lists:member(Entity, Chain) of
                  true -> [{recursive, [Entity|Chain]}|Acc];
                  _    -> [[Entity| Chain]|Acc]
              end
      end,
      [],
      Result).

filter_chain(State, Filter, Chain) ->
    NewChainWithEmpties =
        [ filter(Filter, State#state{res=List}) || List <- Chain ],
    lists:filter( fun(List) -> List /= [] end, NewChainWithEmpties).

%%% ============================================================================
%%% Filters

%% TODO: sort entity lists before filter!!!!
%% @private
%% @spec filter(Filter::term(), #state{}) -> ordset()
filter(_Filter, #state{res=[]}) -> [];

filter('true', #state{res=Entities}) -> Entities;

filter('false', _State) -> [];

filter({'not', Filter}, #state{res=Entities} = State) ->
    ordsets:subtract(Entities, filter(Filter, State));

filter({'or', Filter1, Filter2}, State) ->
    ordsets:union(filter(Filter1, State),
                  filter(Filter2, State));

filter({'and', Filter1, Filter2}, #state{type=EntityType} = State) ->
    filter(Filter2,#state{type= EntityType, res = filter(Filter1, State)});

filter({'in', {seq,SelList}, List},
        #state{params=Params, res=Entities} = State) ->
    ResultList =
        case List of
            {cons,L} -> L;
            {ok,SemQuery} ->
                SemanticQuery = process_semantic_query(Params ,SemQuery),
                [{_,Res},_] = ?Format:result(SemanticQuery, [{output,nodes}]),
                Res
        end,

    ordsets:filter(
      fun(Entity) ->
          #state{type=Type,res = Selected1} =
            process_query_seq(State#state{res = [Entity]},SelList),
          case Selected1 of
            [] -> false;
            _ ->
                lists:map(
                    fun(Any) ->
                        check_filter('in', {element(2,hd(SelList)), Type, Any})
                    end,
                ResultList),

                _Selected1 =
                    case (lists:all(fun is_atom/1,ResultList) andalso
                            Type==string) of
                        true -> lists:map(fun list_to_atom/1,Selected1);
                        _ -> Selected1
                    end,

                _ResultList =
                    case (lists:all(fun is_atom/1,Selected1) andalso
                            lists:all(fun is_list/1,ResultList)) of
                        true -> lists:map(fun list_to_atom/1, ResultList);
                        _ -> ResultList
                    end,
                sets:is_subset(sets:from_list(_Selected1),
                                sets:from_list(_ResultList))
          end
      end,
      Entities);


filter({'in', Property, List},
        #state{params=Params, type=EntityType, res=Entities}) ->
    ResultList =
        case List of
            {cons,L} -> L;
            {'query',Query} ->
                ProcessResult = process_semantic_query(Params,Query),
                [Res||{_,_,_,_,Res}<-?Format:result(ProcessResult,
                                                    [{output,msg}])]

        end,
    lists:map(
        fun(Any) ->
            check_filter('in', {EntityType,
                                hd(?Lib:prop_type(EntityType,Property)), Any})
        end,
        ResultList),
    PropFun = ?Lib:prop_fun(EntityType,Property),
    ordsets:filter(
      fun(Entity) ->
          Result = case PropFun of
                      [] -> Property;
                      [Fun1] -> Fun1(Entity)
                   end,
          case ResultList of
              [] -> false;
              _ ->
                _Result = case (is_atom(Result) andalso
                                lists:all(fun is_list/1,ResultList)) of
                            true -> atom_to_list(Result);
                            _   -> case (lists:all(fun is_atom/1,ResultList)
                                            andalso is_list(Result)) of
                                        true -> list_to_atom(Result);
                                        _    -> Result
                                   end
                    end,
              lists:member(_Result,ResultList)
         end
      end,
      Entities);


filter({seq, QuerySeq}, #state{res=Entities} = State) ->
    ordsets:filter(
      fun(Entity) ->
              St = process_query_seq(State#state{res = [Entity]}, QuerySeq),
                St#state.res /= []
      end,
      Entities);

%% Comparison works on atom, int and string.
%% todo: regexp match -> if a string doesn't match
filter({CompOp, Filt1, Filt2}, #state{type=EntityType, res=Entities}=State) ->
    Comp1 = case Filt1 of
                {quoted, _} -> Filt1;
                _ when is_tuple(Filt1) -> filter(Filt1, State);
                _  -> Filt1
            end,

    Comp2 = case Filt2 of
                {quoted, _} -> Filt2;
                _ when is_tuple(Filt2) -> filter(Filt2, State);
                _ -> Filt2
            end,

    PropFun1 = ?Lib:prop_fun(EntityType, Comp1),
    PropFun2 = ?Lib:prop_fun(EntityType, Comp2),

    ordsets:filter(
      fun(Entity) ->
              CompL = case PropFun1 of [] -> Comp1; [Fun1] -> Fun1(Entity) end,
              CompR = case PropFun2 of [] -> Comp2; [Fun2] -> Fun2(Entity) end,
              compare(CompOp, CompL, CompR)
      end,
      Entities);

filter(Filter, #state{type=EntityType, res=Entities})->
    ordsets:filter(prop_fun(EntityType, Filter), Entities).


%%% ============================================================================
%%% Helper functions

compare(like, CompL, CompR) ->
    Distance =
        if
            is_list(CompL) andalso is_list(CompR) ->
                refusr_strm:getDistance(CompL, CompR);
            is_atom(CompL) andalso is_atom(CompR) ->
                refusr_strm:getDistance(atom_to_list(CompL),
                                        atom_to_list(CompR));
            true -> false
        end,

    case Distance of
         {lev, 0} -> true;
         {lev, 1} -> true;
         {lev, 2} ->
            Length = if is_atom(CompL) -> length(atom_to_list(CompL));
                        true           -> length(CompL)
                     end,
            case Length > 5 of true -> true; _ -> false end;
        _ -> false
    end;

compare('~', CompL, CompR) when is_atom(CompR)->
    compare('~', CompL, atom_to_list(CompR));
compare('~', CompL, CompR) when is_atom(CompL)->
    compare('~', atom_to_list(CompL), CompR);
compare('~', CompL, {quoted, A}) when is_atom(A)->
    compare('~', CompL, atom_to_list(A));
compare('~', {quoted, A}, CompR) when is_atom(A)->
    compare('~', atom_to_list(A), CompR);
compare('~', CompL, CompR) ->
    RegExp = case re:compile(CompR) of
        {ok, MP} -> MP;
        {error, _} -> throw(?LocalErr0r(bad_regexp))
    end,
    case re:run(CompL, RegExp) of
        {match, _} -> true;
        nomatch -> false
    end;


compare(CompOp, {quoted, A}, CompR) ->
    erlang:CompOp(A, CompR);

compare(CompOp, CompL, {quoted, A}) ->
    erlang:CompOp(CompL, A);

compare(CompOp, CompL, CompR) ->
    erlang:CompOp(CompL, CompR).

prop_fun(EntityType, Filter) ->
    case ?Lib:prop_fun(EntityType, Filter) of
        [] -> throw(?LocalError(illegal_property, [EntityType, Filter]));
        [Fun]   -> Fun
    end.

%% @spec format_nodes(Nodes::[entity()], Positions::atom()) -> string()
%%       Positions = none|scalar|linecol
%% @doc Returns a textual representation for a list of nodes.
format_nodes(Nodes, Position) -> ?Format:nodes(Nodes, Position).
